from rest_framework.routers import Route
from rest_framework.routers import SimpleRouter


class AggregationRouter(SimpleRouter):
    """
    A router for read-only APIs, which doesn't use trailing slashes.
    """

    routes = [
        Route(
            url="{prefix}/aggregation",
            mapping={"get": "aggregation"},
            name="{basename}-aggregation",
            detail=False,
            initkwargs={},
        ),
    ]



import json

from django.db import models
from drf_complex_filter.utils import generate_query_from_dict
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet

from .aggregates import Percentile
from .utils import get_aggregation


class AggregationViewSet(GenericViewSet):
    def aggregation(self, request):
        annotation = self._get_annotation(request=request)

        result = get_aggregation(
            queryset=self.filter_queryset(self.get_queryset()),
            annotation=annotation,
            group_by=self._get_group_by(request=request),
            order_by=self._get_order(request=request),
            limit=self._get_limit(request=request),
            show_other=self._get_show_other(request=request),
        )

        return Response(result)

    def _get_annotation(self, request):
        aggregation = self._get_aggregation(request)
        if aggregation == 'count':
            return models.Count('id')

        if aggregation == 'sum':
            aggregation_field = self._get_aggregation_field(request=request)
            return models.Sum(aggregation_field)

        if aggregation == 'average':
            aggregation_field = self._get_aggregation_field(request=request)
            return models.Avg(aggregation_field)

        if aggregation == 'minimum':
            aggregation_field = self._get_aggregation_field(request=request)
            return models.Min(aggregation_field)

        if aggregation == 'maximum':
            aggregation_field = self._get_aggregation_field(request=request)
            return models.Max(aggregation_field)

        if aggregation == 'percentile':
            aggregation_field = self._get_aggregation_field(request=request)
            percentile = self._get_percentile(request)
            return Percentile(aggregation_field, percentile)

        if aggregation == 'percent':
            raise NotImplementedError("Grouped percentage not yet implemented")

    @staticmethod
    def _get_aggregation(request) -> str:
        aggregation = request.query_params.get("aggregation", None)
        if not aggregation:
            raise ValidationError({"error": "Aggregation is mandatory."})

        return aggregation

    @staticmethod
    def _get_group_by(request) -> list:
        group_by = request.query_params.get("groupBy", None)
        group_by = group_by.split(",") if group_by else []

        return group_by

    @staticmethod
    def _get_order(request) -> list:
        order = request.query_params.get("order", None)
        if order == "asc":
            return ["value"]
        if order == "desc":
            return ["-value"]

        return []

    @staticmethod
    def _get_limit(request) -> (int, None):
        limit = request.query_params.get("limit", None)
        return int(limit) if limit else None

    @staticmethod
    def _get_show_other(request) -> bool:
        show_other = request.query_params.get("showOther", None)
        return show_other == "1"

    @staticmethod
    def _get_aggregation_field(request) -> str:
        aggregation_field = request.query_params.get("aggregationField", None)
        if not aggregation_field:
            raise ValidationError({"error": "Aggregation field is mandatory."})

        return aggregation_field

    @staticmethod
    def _get_additional_query(request) -> models.Q:
        try:
            additional_filter = json.loads(
                request.query_params.get("additionalFilter", None)
            )
        except (TypeError, json.decoder.JSONDecodeError):
            raise ValidationError({"error": "Additional filter is mandatory."})
        additional_query = generate_query_from_dict(additional_filter)
        if not additional_query:
            raise ValidationError(
                {"error": "Additional filter cannot be empty."}
            )

        return additional_query

    @staticmethod
    def _get_percentile(request) -> str:
        percentile = request.query_params.get("percentile", None)
        if not percentile:
            raise ValidationError({"error": "Percentile is mandatory."})

        return percentile



from django.db import models

OTHER_GROUP_NAME = "Other"


def get_aggregation(
        queryset,
        annotation,
        group_by: list = None,
        order_by: list = None,
        limit: int = None,
        show_other: bool = False
) -> (dict, list):
    """Get the sum of the values of an aggregation field

    :param queryset: Django Queryset for aggregation
    :param annotation:
    :param group_by: list of fields to group the result
    :param order_by: list of fields to sort the result
    :param limit: number of record to return
    :param show_other: if a limit is set,
        combine other records into an additional group
    :return: Django Queryset with aggregated data
    """
    if not group_by:
        return queryset.aggregate(value=annotation)

    aggregation = queryset.values(*group_by)
    aggregation = aggregation.annotate(value=annotation)
    if order_by:
        aggregation = aggregation.order_by(*order_by)
    if not limit:
        return list(aggregation)

    aggregation = list(aggregation[: limit])
    if not show_other:
        return aggregation

    main_group_name = group_by[0]
    values = {result[main_group_name] for result in aggregation}
    additional_queryset = _get_queryset_without_groups(
        queryset,
        main_group_name,
        values
    )
    additional_aggregation = get_aggregation(
        queryset=additional_queryset,
        annotation=annotation,
        group_by=group_by[1:]
    )
    aggregation = _merge_aggregations(
        aggregation=aggregation,
        additional_aggregation=additional_aggregation,
        field_name=main_group_name,
    )
    return aggregation


def _get_queryset_without_groups(queryset: models.QuerySet,
                                 field_name: str,
                                 values: set) -> models.QuerySet:
    queryset = queryset.all()
    for value in values:
        query = ~models.Q(**{"{}".format(field_name): value})
        queryset = queryset.filter(query)

    return queryset


def _merge_aggregations(
        aggregation: list,
        additional_aggregation: (dict, list),
        field_name: str
) -> list:
    merged_aggregation = list(aggregation)
    if isinstance(additional_aggregation, dict):
        additional_aggregation[field_name] = OTHER_GROUP_NAME
        merged_aggregation.append(additional_aggregation)
    elif isinstance(additional_aggregation, list):
        for result in additional_aggregation:
            result[field_name] = OTHER_GROUP_NAME
            merged_aggregation.append(result)

    return merged_aggregation



""" From Django Postgres Stats
https://github.com/rtidatascience/django-postgres-stats
"""
from django.db.models import Aggregate


class Percentile(Aggregate):
    """
    Accepts a numerical field or expression and a list of fractions and
    returns values for each fraction given corresponding to that fraction in
    that expression.

    If *continuous* is True (the default), the value will be interpolated
    between adjacent values if needed. Otherwise, the value will be the first
    input value whose position in the ordering equals or exceeds the
    specified fraction.

    You will likely have to declare the *output_field* for your results.
    Django cannot guess what type of value will be returned.

    Usage example::

        from django.contrib.postgres.fields import ArrayField

        numbers = [31, 83, 237, 250, 305, 314, 439, 500, 520, 526, 527, 533,
                   540, 612, 831, 854, 857, 904, 928, 973]
        for n in numbers:
            Number.objects.create(n=n)

        results = Number.objects.all().aggregate(
            median=Percentile('n', 0.5, output_field=models.FloatField()))
        assert results['median'] == 526.5

        results = Number.objects.all().aggregate(
            quartiles=Percentile('n', [0.25, 0.5, 0.75],
                                 output_field=ArrayField(models.FloatField())))
        assert results['quartiles'] == [311.75, 526.5, 836.75]

        results = Number.objects.all().aggregate(
            quartiles=Percentile('n', [0.25, 0.5, 0.75],
                                 continuous=False,
                                 output_field=ArrayField(models.FloatField())))
        assert results['quartiles'] == [305, 526, 831]
    """

    function = None
    name = "percentile"
    template = "%(function)s(%(percentiles)s) WITHIN GROUP (ORDER BY %(" \
               "expressions)s)"

    def __init__(self, expression, percentiles, continuous=True, **extra):
        if isinstance(percentiles, (list, tuple)):
            percentiles = "array%(percentiles)s" % {'percentiles': percentiles}
        if continuous:
            extra['function'] = 'PERCENTILE_CONT'
        else:
            extra['function'] = 'PERCENTILE_DISC'

        super().__init__(expression, percentiles=percentiles, **extra)
